!===============================================================================
!
! cg_d25.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2016 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# define LANGUAGE_FORTRAN
# include "defs.h"
# include "cg_defs.h"

!-------------------------------------------------------------------------------
#ifdef DAGGER
# define NAME cg_d25_dag
# define NAME2 cg_d2_dag
#else
# define NAME cg_d25
# define NAME2 cg_d2
#endif

# define TRACE_NAME T(NAME)
!-------------------------------------------------------------------------------
subroutine T(NAME)(e, o, out, in, u, a)

!// out is of type "e" = EVEN or ODD
!// in is of type "o" = ODD or EVEN

  use typedef_simd
  use T(module_nn)
  use T(module_vol)
  implicit none
 
  integer, intent(in)          :: e, o
  spincol_FIELD, intent(out)   :: out
  spincol_FIELD, intent(inout) :: in   ! 'xbound' modifies 'in'
  g1_FIELD, intent(in)         :: u
  sc2_FIELD, intent(out)       :: a    ! scratch array

#ifdef DAGGER
  integer, parameter :: dag = 1
#else
  integer, parameter :: dag = 0
#endif

  TRACE_ENTRY
  TIMING_START(timing_bin_cg_d)

  call T(cg_proj)(a, in, 0, dag)
  call T(cg_proj_xbound)(in, a, 0, dag)
  call cg_proj_wait(0)

  call T(STRCAT(NAME2, _t1))(out, in, u(1, 1, 1, e, 4), u(1, 1, 1, o, 4), &
                                           nn(1, e, 4, FWD), nn(1, e, 4, BWD), volh_tot, volh)
  call T(STRCAT(NAME2, _z))(out, in, u(1, 1, 1, e, 3), u(1, 1, 1, o, 3), &
                                          nn(1, e, 3, FWD), nn(1, e, 3, BWD), volh_tot, volh)
  call T(STRCAT(NAME2, _y))(out, in, u(1, 1, 1, e, 2), u(1, 1, 1, o, 2), &
                                          nn(1, e, 2, FWD), nn(1, e, 2, BWD), volh_tot, volh)

#ifdef SIMD
  call T(STRCAT(NAME2, _x))(out, in, u(1, 1, 1, e, 1), u(1, 1, 1, e, 5), &
#else
  call T(STRCAT(NAME2, _x))(out, in, u(1, 1, 1, e, 1), u(1, 1, 1, o, 1), &
#endif
                                          nn(1, e, 1, FWD), nn(1, e, 1, BWD), volh_tot, volh)

  TIMING_STOP(timing_bin_cg_d)
  TRACE_RETURN
end

!===============================================================================
