include(`defs.m4')
include(`D104_defs.m4')
include(`spin_proj.m4')
include(`spin_sum_inner.m4')
!===============================================================================
!
! D104.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2012 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine emuu4();end
subroutine d_xbound();end
subroutine d_dag_projection_block();end
subroutine d_projection_block();end
subroutine d_dag_x_block();end
subroutine d_dag_y_block();end
subroutine d_dag_z_block();end
subroutine d_dag_t_block();end
subroutine d_x_block();end
subroutine d_y_block();end
subroutine d_z_block();end
subroutine d_t_block();end

#ifndef PRECISION_R4
subroutine apply_mp_chemu();end
subroutine init_d();end
!-------------------------------------------------------------------------------
subroutine d_g_init(u)
  use module_vol
  implicit none
  GAUGE_FIELD :: u

  return

end
!-------------------------------------------------------------------------------
#endif
!===============================================================================
!-------------------------------------------------------------------------------
subroutine d(e, o, out, in, u)
  use module_vol
  use module_nn
  implicit none

  integer :: e, o
  COMPLEX, dimension (4, 3, volh_tot) :: out, in
  COMPLEX, dimension (3, 3, volh_tot, 0:1, 4) :: u
  COMPLEX :: u11, u21, u31, u12, u22, u32, u13, u23, u33, &
             a11, a21, a12, a22, a13, a23, &
             b11, b21, b12, b22, b13, b23, &
             bb11, bb21, bb31, bb41, &
             bb12, bb22, bb32, bb42, &
             bb13, bb23, bb33, bb43
  integer :: i, ixf, ixb, iyf, iyb, izf, izb, itf, itb

  call timing_start(12)

  !$omp parallel do private(ixf, ixb, iyf, iyb, izf, izb, itf, itb, &
  !$omp u11, u21, u31, u12, u22, u32, u13, u23, u33, &
  !$omp a11, a21, a12, a22, a13, a23, &
  !$omp b11, b21, b12, b22, b13, b23, &
  !$omp bb11, bb21, bb31, bb41, &
  !$omp bb12, bb22, bb32, bb42, &
  !$omp bb13, bb23, bb33, bb43)
  do i = 1, volh
     ixf = nn(i,e,1,FWD)
     ixb = nn(i,e,1,BWD)
     iyf = nn(i,e,2,FWD)
     iyb = nn(i,e,2,BWD)
     izf = nn(i,e,3,FWD)
     izb = nn(i,e,3,BWD)
     itf = nn(i,e,4,FWD)
     itb = nn(i,e,4,BWD)

     bb11 = ZERO
     bb21 = ZERO
     bb31 = ZERO
     bb41 = ZERO
     bb12 = ZERO
     bb22 = ZERO
     bb32 = ZERO
     bb42 = ZERO
     bb13 = ZERO
     bb23 = ZERO
     bb33 = ZERO
     bb43 = ZERO

  if (ixf <= volh ) then
_load_su3(u, i, 1, e)
_proj_1_p_gamma1(a, in, `,ixf')
_mult_spinor_su3(b, a)
_sum_1_p_gamma1(bb, b, 1)
_sum_1_p_gamma1(bb, b, 2)
_sum_1_p_gamma1(bb, b, 3)
  endif

  if (ixb <= volh ) then
_load_su3dag(u, ixb, 1, o)
_proj_1_m_gamma1(a, in, `,ixb')
_mult_spinor_su3(b, a)
_sum_1_m_gamma1(bb, b, 1)
_sum_1_m_gamma1(bb, b, 2)
_sum_1_m_gamma1(bb, b, 3)
  endif

  if (iyf <= volh ) then
_load_su3(u, i, 2, e)
_proj_1_p_gamma2(a, in, `,iyf')
_mult_spinor_su3(b, a)
_sum_1_p_gamma2(bb, b, 1)
_sum_1_p_gamma2(bb, b, 2)
_sum_1_p_gamma2(bb, b, 3)
  endif

  if (iyb <= volh ) then
_load_su3dag(u, iyb, 2, o)
_proj_1_m_gamma2(a, in, `,iyb')
_mult_spinor_su3(b, a)
_sum_1_m_gamma2(bb, b, 1)
_sum_1_m_gamma2(bb, b, 2)
_sum_1_m_gamma2(bb, b, 3)
  endif

  if (izf <= volh ) then
_load_su3(u, i, 3, e)
_proj_1_p_gamma3(a, in, `,izf')
_mult_spinor_su3(b, a)
_sum_1_p_gamma3(bb, b, 1)
_sum_1_p_gamma3(bb, b, 2)
_sum_1_p_gamma3(bb, b, 3)
  endif

  if (izb <= volh ) then
_load_su3dag(u, izb, 3, o)
_proj_1_m_gamma3(a, in, `,izb')
_mult_spinor_su3(b, a)
_sum_1_m_gamma3(bb, b, 1)
_sum_1_m_gamma3(bb, b, 2)
_sum_1_m_gamma3(bb, b, 3)
  endif

  if (itf <= volh ) then
_load_su3(u, i, 4, e)
_proj_1_m_gamman(a, in, `,itf')
_mult_spinor_su3(b, a)
_sum_1_m_gamman(bb, b, 1)
_sum_1_m_gamman(bb, b, 2)
_sum_1_m_gamman(bb, b, 3)
  endif

  if (itb <= volh ) then
_load_su3dag(u, itb, 4, o)
_proj_1_p_gamman(a, in, `,itb')
_mult_spinor_su3(b, a)
_sum_1_p_gamman(bb, b, 1)
_sum_1_p_gamman(bb, b, 2)
_sum_1_p_gamman(bb, b, 3)
  endif

     out(1, 1, i) = bb11
     out(2, 1, i) = bb21
     out(3, 1, i) = bb31
     out(4, 1, i) = bb41
     out(1, 2, i) = bb12
     out(2, 2, i) = bb22
     out(3, 2, i) = bb32
     out(4, 2, i) = bb42
     out(1, 3, i) = bb13
     out(2, 3, i) = bb23
     out(3, 3, i) = bb33
     out(4, 3, i) = bb43

  enddo

  call timing_stop(12)
end

!-------------------------------------------------------------------------------
subroutine d_dag(e, o, out, in, u)
  use module_vol
  use module_nn
  implicit none

  integer :: e, o
  COMPLEX, dimension (4, 3, volh_tot) :: out, in
  COMPLEX, dimension (3, 3, volh_tot, 0:1, 4) :: u
  COMPLEX :: u11, u21, u31, u12, u22, u32, u13, u23, u33, &
             a11, a21, a12, a22, a13, a23, &
             b11, b21, b12, b22, b13, b23, &
             bb11, bb21, bb31, bb41, &
             bb12, bb22, bb32, bb42, &
             bb13, bb23, bb33, bb43
  integer :: i, ixf, ixb, iyf, iyb, izf, izb, itf, itb

  call timing_start(12)

  !$omp parallel do private(ixf, ixb, iyf, iyb, izf, izb, itf, itb, &
  !$omp u11, u21, u31, u12, u22, u32, u13, u23, u33, &
  !$omp a11, a21, a12, a22, a13, a23, &
  !$omp b11, b21, b12, b22, b13, b23, &
  !$omp bb11, bb21, bb31, bb41, &
  !$omp bb12, bb22, bb32, bb42, &
  !$omp bb13, bb23, bb33, bb43)
  do i = 1, volh
     ixf = nn(i,e,1,FWD)
     ixb = nn(i,e,1,BWD)
     iyf = nn(i,e,2,FWD)
     iyb = nn(i,e,2,BWD)
     izf = nn(i,e,3,FWD)
     izb = nn(i,e,3,BWD)
     itf = nn(i,e,4,FWD)
     itb = nn(i,e,4,BWD)

     bb11 = ZERO
     bb21 = ZERO
     bb31 = ZERO
     bb41 = ZERO
     bb12 = ZERO
     bb22 = ZERO
     bb32 = ZERO
     bb42 = ZERO
     bb13 = ZERO
     bb23 = ZERO
     bb33 = ZERO
     bb43 = ZERO

  if (ixf <= volh ) then
_load_su3(u, i, 1, e)
_proj_1_m_gamma1(a, in, `,ixf')
_mult_spinor_su3(b, a)
_sum_1_m_gamma1(bb, b, 1)
_sum_1_m_gamma1(bb, b, 2)
_sum_1_m_gamma1(bb, b, 3)
  endif

  if (ixb <= volh ) then
_load_su3dag(u, ixb, 1, o)
_proj_1_p_gamma1(a, in, `,ixb')
_mult_spinor_su3(b, a)
_sum_1_p_gamma1(bb, b, 1)
_sum_1_p_gamma1(bb, b, 2)
_sum_1_p_gamma1(bb, b, 3)
  endif

  if (iyf <= volh ) then
_load_su3(u, i, 2, e)
_proj_1_m_gamma2(a, in, `,iyf')
_mult_spinor_su3(b, a)
_sum_1_m_gamma2(bb, b, 1)
_sum_1_m_gamma2(bb, b, 2)
_sum_1_m_gamma2(bb, b, 3)
  endif

  if (iyb <= volh ) then
_load_su3dag(u, iyb, 2, o)
_proj_1_p_gamma2(a, in, `,iyb')
_mult_spinor_su3(b, a)
_sum_1_p_gamma2(bb, b, 1)
_sum_1_p_gamma2(bb, b, 2)
_sum_1_p_gamma2(bb, b, 3)
  endif

  if (izf <= volh ) then
_load_su3(u, i, 3, e)
_proj_1_m_gamma3(a, in, `,izf')
_mult_spinor_su3(b, a)
_sum_1_m_gamma3(bb, b, 1)
_sum_1_m_gamma3(bb, b, 2)
_sum_1_m_gamma3(bb, b, 3)
  endif

  if (izb <= volh ) then
_load_su3dag(u, izb, 3, o)
_proj_1_p_gamma3(a, in, `,izb')
_mult_spinor_su3(b, a)
_sum_1_p_gamma3(bb, b, 1)
_sum_1_p_gamma3(bb, b, 2)
_sum_1_p_gamma3(bb, b, 3)
  endif

  if (itf <= volh ) then
_load_su3(u, i, 4, e)
_proj_1_p_gamman(a, in, `,itf')
_mult_spinor_su3(b, a)
_sum_1_p_gamman(bb, b, 1)
_sum_1_p_gamman(bb, b, 2)
_sum_1_p_gamman(bb, b, 3)
  endif

  if (itb <= volh ) then
_load_su3dag(u, itb, 4, o)
_proj_1_m_gamman(a, in, `,itb')
_mult_spinor_su3(b, a)
_sum_1_m_gamman(bb, b, 1)
_sum_1_m_gamman(bb, b, 2)
_sum_1_m_gamman(bb, b, 3)
  endif

     out(1, 1, i) = bb11
     out(2, 1, i) = bb21
     out(3, 1, i) = bb31
     out(4, 1, i) = bb41
     out(1, 2, i) = bb12
     out(2, 2, i) = bb22
     out(3, 2, i) = bb32
     out(4, 2, i) = bb42
     out(1, 3, i) = bb13
     out(2, 3, i) = bb23
     out(3, 3, i) = bb33
     out(4, 3, i) = bb43

  enddo

  call timing_stop(12)

end
!-------------------------------------------------------------------------------
