#===============================================================================
#
# d/libd103.mk - libd100 with SSE for GAMMA_NOTATION=BQCD
#
#-------------------------------------------------------------------------------
#
# Copyright (C) 2008-2011 Yoshifumi Nakamura, Hinnerk Stueben
#
# This file is part of BQCD -- Berlin Quantum ChromoDynamics program
#
# BQCD is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BQCD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================

FAST_MAKE = make -j 1

M4=m4

MODULES = \
	module_d21.o \
	module_d21_r4.o \

ifdef quad
MODULES += module_d21_r16.o
endif

OBJS = \
	d100_g.o \
	d103_version.o \
	d103.o \
	dxyzt.o \
	dxyzt_block.o \
	dxyzt_block2.o \
	dxyzt_r4.o \
	d103_r4.o \
	dxyzt_block_r4.o \
	dxyzt_block2_r4.o \
	chemicalp.o chemicalp_r4.o

ifdef quad
OBJS +=	dxyzt_r16.o \
	d100_r16.o \
	dxyzt_block_r16.o
endif


R4_FLAGS = -DPRECISION_R4 -DD_R4_INTERNAL

fast:
	$(MAKE) $(MODULES)
	$(FAST_MAKE)

module_d21.f90:    module_d21.F90;  $(fpp) $< > $@
module_d21_r4.f90: module_d21.F90;  $(fpp) -DPRECISION_R4 $< > $@
module_d21_r16.f90: module_d21.F90;  $(fpp) -DPRECISION_R16 $< > $@

d100.F90:       D100.m4;       $(M4) -I../../include $(MYFLAGS) D100.m4       >d100.F90
d103.F90:       D103.m4;       $(M4) -I../../include $(MYFLAGS) D103.m4       >d103.F90

dxyzt.c: D103.c
	$(CPP) -I../../include $(CFLAGS_STD) $(MYFLAGS) $< > $@

dxyzt_r4.c: D103_r4.c
	$(CPP) -I../../include $(CFLAGS_STD) $(MYFLAGS) $< > $@

dxyzt_block2.c: D103_block.c
	$(CPP) -I../../include $(CFLAGS_STD) $(MYFLAGS) $< > $@

dxyzt_block2_r4.c: D103_block_r4.c
	$(CPP) -I../../include $(CFLAGS_STD) $(MYFLAGS) $< > $@

dxyzt.c: D103xyzt.c
dxyzt_r4.c: D103xyzt_r4.c
dxyzt_block2.c: D103xyzt.c
dxyzt_block2_r4.c: D103xyzt_r4.c

dxyzt.F90: D100_util.m4
	$(M4) -I../../include $(MYFLAGS) -DNAME=xf     -DDIRECTION=X          D100_util.m4 > dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=yf     -DDIRECTION=Y          D100_util.m4 >>dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=zf     -DDIRECTION=Z          D100_util.m4 >>dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=t      -DDIRECTION=T          D100_util.m4 >>dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_xf -DDIRECTION=X -DDAGGER D100_util.m4 >>dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_yf -DDIRECTION=Y -DDAGGER D100_util.m4 >>dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_zf -DDIRECTION=Z -DDAGGER D100_util.m4 >>dxyzt.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_t  -DDIRECTION=T -DDAGGER D100_util.m4 >>dxyzt.F90

dxyzt_block.F90: D103_util_block.m4
	$(M4) -I../../include $(MYFLAGS) -DNAME=x     -DDIRECTION=X          D103_util_block.m4 > dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=y     -DDIRECTION=Y          D103_util_block.m4 >>dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=z     -DDIRECTION=Z          D103_util_block.m4 >>dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=t     -DDIRECTION=T          D103_util_block.m4 >>dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_x -DDIRECTION=X -DDAGGER D103_util_block.m4 >>dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_y -DDIRECTION=Y -DDAGGER D103_util_block.m4 >>dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_z -DDIRECTION=Z -DDAGGER D103_util_block.m4 >>dxyzt_block.F90
	$(M4) -I../../include $(MYFLAGS) -DNAME=dag_t -DDIRECTION=T -DDAGGER D103_util_block.m4 >>dxyzt_block.F90

d100_g.f90:       D100_g.F90;	$(fpp) D100_g.F90 > $@
d103_version.f90: DVersion.F90; $(fpp) -DVERSION=103 DVersion.F90 > $@

d103.f90: d103.F90;	                $(fpp) d103.F90 > d103.f90
dxyzt_block.f90: dxyzt_block.F90;	$(fpp) dxyzt_block.F90 > $@

d103_r4.f90: d103.F90;	                $(fpp) -DPRECISION_R4 $< > $@
dxyzt_block_r4.f90: dxyzt_block.F90;	$(fpp) -DPRECISION_R4 $< > $@

d100_r16.f90: d100.F90;	                $(fpp) -DPRECISION_R16 $< > $@
dxyzt_r16.f90: dxyzt.F90;	        $(fpp) -DPRECISION_R16 $< > $@
dxyzt_block_r16.f90: dxyzt_block.F90;	$(fpp) -DPRECISION_R16 $< > $@

chemicalp.f90:    chemicalp.F90;  $(fpp) $< > $@
chemicalp_r4.f90:    chemicalp.F90;  $(fpp) -DPRECISION_R4 $< > $@
