/*
 * storing one color vector in two
 * double4vectors.
 *
 * storing the su3 matrix in
 * 3 color vectors (6 double4vectors)
 *   cv1 = a[0] = c11 c22
 *         a[1] = c33
 *   cv2 = a[2] = c31 c12
 *         a[3] = c23
 *   cv3 = a[4] = c21 c32
 *         a[5] = c13
 *
 * */
#define _qpx_vector_load(a,b) \
    a[0] = vec_ld(0l,&b.c1.re); \
    a[1] = vec_ld2(16l,&b.c3.re) 

#define _qpx_vector_add(a,b,c) \
    a[0] = vec_add(b[0],c[0]); \
    a[1] = vec_add(b[1],c[1])

#define _qpx_su3_load(a,b) \
    a[0] = vecld(0l,&b.c11.re);     /* c11 (c12) */ \
    a[5] = vecld(0l,&b.c13.re);     /* c13 (c21) */ \
    w = vecld(0l,&b.c22.re);        /* (c22 c23) */ \
    a[2] = vecld(0,&(b.c31.re);     /* c31 (c32) */ \
    a[1] = vecld2(0,&(b.c33.re);    /* c33 xx */ \
        \
    perm = vec_gpci( 02367 ); \
    a[3] = vec_perm( w, w, perm );    /* c23 */ \
    a[4] = vec_perm( a[5], a[2], perm );        /* c21 c32 */ \
        \
    perm = vec_gpci( 00167 ); \
    a[2] = vec_perm( a[2], a[0], perm );        /* c31 c12 */ \
        \
    perm = vec_gpci( 00145 ); \
    a[0] = vec_perm( a[0], w, perm )       /* c11 c22 */ \

#define _qpx_su3_multiply(ra,u,a) \
    ra[0] = vec_xmul( a[0], u[0] ); \
    ra[1] = vec_xmul( a[1], u[1] ); \
        \
    ra[0] = vec_xxnpmadd( u[0], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[1], a[1], ra[1] ); \
        \
    ra[0] = vec_xmadd( a[0], u[2], ra[0] ); \
    ra[1] = vec_xmadd( a[1], u[3], ra[1] ); \
        \
    ra[0] = vec_xxnpmadd( u[2], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[3], a[1], ra[1] ); \
        \
    ra[0] = vec_xmadd( a[0], u[4], ra[0] ); \
    ra[1] = vec_xmadd( a[1], u[5], ra[1] ); \
        \
    ra[0] = vec_xxnpmadd( u[4], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[5], a[1], ra[1] )

#define _qpx_su3_multiply2(ra,rb,u,a,b) \
    ra[0] = vec_xmul( a[0], u[0] ); \
    ra[1] = vec_xmul( a[1], u[1] ); \
    rb[0] = vec_xmul( b[0], u[2] ); \
    rb[1] = vec_xmul( b[1], u[3] ); \
        \
    ra[0] = vec_xxnpmadd( u[0], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[1], a[1], ra[1] ); \
    rb[0] = vec_xxnpmadd( u[2], a[0], rb[0] ); \
    rb[1] = vec_xxnpmadd( u[3], a[1], rb[1] ); \
        \
    ra[0] = vec_xmadd( a[0], u[2], ra[0] ); \
    ra[1] = vec_xmadd( a[1], u[3], ra[1] ); \
    rb[0] = vec_xmadd( b[0], u[4], rb[0] ); \
    rb[1] = vec_xmadd( b[1], u[5], rb[1] ); \
        \
    ra[0] = vec_xxnpmadd( u[2], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[3], a[1], ra[1] ); \
    rb[0] = vec_xxnpmadd( u[4], a[0], rb[0] ); \
    rb[1] = vec_xxnpmadd( u[5], a[1], rb[1] ); \
        \
    ra[0] = vec_xmadd( a[0], u[4], ra[0] ); \
    ra[1] = vec_xmadd( a[1], u[5], ra[1] ); \
    rb[0] = vec_xmadd( b[0], u[0], rb[0] ); \
    rb[1] = vec_xmadd( b[1], u[1], rb[1] ); \
        \
    ra[0] = vec_xxnpmadd( u[4], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[5], a[1], ra[1] ); \
    rb[0] = vec_xxnpmadd( u[0], a[0], rb[0] ); \
    rb[1] = vec_xxnpmadd( u[1], a[1], rb[1] ) 
    
#define _qpx_su3_inv_multiply(ra,u,a) \
    ra[0] = vec_xmul( a[0], u[0] ); \
    ra[1] = vec_xmul( a[1], u[1] ); \
        \
    ra[0] = vec_mul( ra[0], perm ); \
    ra[1] = vec_mul( ra[1], perm ); \
        \
    ra[0] = vec_xxnpmadd( u[0], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[1], a[1], ra[1] ); \
        \
    ra[0] = vec_xmadd( a[0], u[2], ra[0] ); \
    ra[1] = vec_xmadd( a[1], u[3], ra[1] ); \
        \
    ra[0] = vec_mul( ra[0], perm ); \
    ra[1] = vec_mul( ra[1], perm ); \
        \
    ra[0] = vec_xxnpmadd( u[2], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[3], a[1], ra[1] ); \
        \
    ra[0] = vec_xmadd( a[0], u[4], ra[0] ); \
    ra[1] = vec_xmadd( a[1], u[5], ra[1] ); \
        \
    ra[0] = vec_mul( ra[0], perm ); \
    ra[1] = vec_mul( ra[1], perm ); \
        \
    ra[0] = vec_xxnpmadd( u[4], a[0], ra[0] ); \
    ra[1] = vec_xxnpmadd( u[5], a[1], ra[1] )

