/* 
 * store one color component of
 * a half spinor in a double4vector
 * both "independent" spincomponents
 * are then stored in 3 double4vectors p0,p1,p2
 *
 * store su3 matrix components (i,j)
 * in separate double4vectors
 * v_ij = { re_ij, im_ij, re_ij, im_ij }
 * maybe loading one column or row at a
 * time is sufficient
 *
 * */

#ifndef QPX_H
#define QPX_H

#warning "Using new QPX intrinsics"

#if 1
#define VEC_LOAD vec_lda
#define VEC_LOAD2 vec_ld2a
#define VEC_LOAD_S vec_lds
#define VEC_STORE vec_sta
#define VEC_STORE2 vec_st2a
#else
#define VEC_LOAD vec_ld
#define VEC_LOAD2 vec_ld2
#define VEC_LOAD_S vec_lds
#define VEC_STORE vec_st
#define VEC_STORE2 vec_st2
#endif

/*
 * vector functions
 *
 * */

#define _qpx_raw_spinor_declare(v) \
  vector4double v##_0, v##_1, v##_2, v##_3, v##_4, v##_5 



#define _qpx_raw_spinor_load(v, spinor) \
  v##_0 = VEC_LOAD(0l, &##spinor##.c1.c1.re); \
  v##_1 = VEC_LOAD(0l, &##spinor##.c1.c3.re); \
  v##_2 = VEC_LOAD(0l, &##spinor##.c2.c2.re); \
  v##_3 = VEC_LOAD(0l, &##spinor##.c3.c1.re); \
  v##_4 = VEC_LOAD(0l, &##spinor##.c3.c3.re); \
  v##_5 = VEC_LOAD(0l, &##spinor##.c4.c2.re); \
  

#define _qpx_raw_spinor_store(v, spinor) \
  VEC_STORE(v##_0, 0l, &##spinor##.c1.c1.re);  \
  VEC_STORE(v##_1, 0l, &##spinor##.c1.c3.re);  \
  VEC_STORE(v##_2, 0l, &##spinor##.c2.c2.re);  \
  VEC_STORE(v##_3, 0l, &##spinor##.c3.c1.re);  \
  VEC_STORE(v##_4, 0l, &##spinor##.c3.c3.re);  \
  VEC_STORE(v##_5, 0l, &##spinor##.c4.c2.re);  \
  
#define _qpx_dcomplex_load(v, cval)		\
  v = VEC_LOAD2(0l, (double*) &(cval))

// r = a*b+c 
#define _qpx_complex_madd(r, a, b, c)		    \
  r = vec_xxnpmadd( a, b, vec_xmadd(b, a, c ) )    \


  //vio = vio + c*vin
#define _qpx_raw_spinor_mcsa(vio, vin, c) \
  _qpx_complex_madd(vio##_0, c, vin##_0, vio##_0); \
  _qpx_complex_madd(vio##_1, c, vin##_1, vio##_1); \
  _qpx_complex_madd(vio##_2, c, vin##_2, vio##_2); \
  _qpx_complex_madd(vio##_3, c, vin##_3, vio##_3); \
  _qpx_complex_madd(vio##_4, c, vin##_4, vio##_4); \
  _qpx_complex_madd(vio##_5, c, vin##_5, vio##_5); \


#define _qpx_vector_declare(v) \
    vector4double v

#define _qpx_vectorpair_declare(p) \
    vector4double p##_0, p##_1, p##_2 

#define _qpx_vectorpair_static_declare(p) \
    static vector4double p##_0, p##_1, p##_2

/*  print   */

#define _qpx_vector_print(a) \
    printf( "dbl4vec = {%f, %f, %f, %f}\n", a[0], a[1], a[2], a[3] )

#define _qpx_vectorpair_print(p) \
    printf( "vec1 = {%f+j%f, %f+j%f, %f+j%f}\n", p##_0[0], p##_0[1], p##_1[0], p##_1[1], p##_2[0], p##_2[1] ); \
    printf( "vec2 = {%f+j%f, %f+j%f, %f+j%f}\n", p##_0[2], p##_0[3], p##_1[2], p##_1[3], p##_2[2], p##_2[3] ) 

/*  loads   */

#define _qpx_vectorpair_verbose_load(p,a,b,w,perm) \
    p##_2 = VEC_LOAD(0l,&a.c1.re); \
    w = VEC_LOAD(0l,&b.c1.re); \
    _qpx_vector_print( p##_2 ); \
    _qpx_vector_print( w ); \
        \
    perm = vec_gpci( 02367 ); \
    p##_1 = vec_perm( p##_2, w, perm ); /* a.c2 b.c2 */ \
    _qpx_vector_print( p##_1 ); \
    perm = vec_gpci( 00145 ); \
    p##_0 = vec_perm( p##_2, w, perm ); /* a.c1 b.c1 */ \
        \
    p##_2 = VEC_LOAD2(0l,&a.c3.re); \
    w = VEC_LOAD2(0l,&b.c3.re); \
    _qpx_vector_print( p##_2 ); \
    _qpx_vector_print( w ); \
    p##_2 = vec_perm( p##_2, w, perm ); \
    _qpx_vector_print( p##_2 ) 

#define _qpx_vectorpair_aa_load(p,a,b,w,perm) \
    p##_2 = VEC_LOAD(0l,&a.c1.re); \
    w = VEC_LOAD(0l,&b.c1.re); \
        \
    perm = vec_gpci( 02367 ); \
    p##_1 = vec_perm( p##_2, w, perm ); /* a.c2 b.c2 */ \
    perm = vec_gpci( 00145 ); \
    p##_0 = vec_perm( p##_2, w, perm ); /* a.c1 b.c1 */ \
        \
    p##_2 = VEC_LOAD2(0l,&a.c3.re); \
    w = VEC_LOAD2(0l,&b.c3.re); \
    p##_2 = vec_perm( p##_2, w, perm ) /* a.c3 b.c3 */ 

#define _qpx_vectorpair_uu_load(p,a,b,w,perm) \
    p##_0 = VEC_LOAD(0l,&a.c2.re); \
    w = VEC_LOAD(0l,&b.c2.re); \
        \
    perm = vec_gpci( 02367 ); \
    p##_2 = vec_perm( p##_0, w, perm ); /* a.c3 b.c3 */ \
    perm = vec_gpci( 00145 ); \
    p##_1 = vec_perm( p##_0, w, perm ); /* a.c2 b.c2 */ \
        \
    p##_0 = VEC_LOAD2(0l,&a.c1.re); \
    w = VEC_LOAD2(0l,&b.c1.re); \
    p##_0 = vec_perm( p##_0, w, perm ) /* a.c1 a.c1 */ 

#define _qpx_vectorpair_au_load(p,a,b,w,perm) \
    p##_0 = VEC_LOAD(0l,&a.c1.re); \
    p##_2 = VEC_LOAD2(0l,&b.c1.re); \
    w = VEC_LOAD(0l,&b.c2.re); \
        \
    perm = vec_gpci( 02345 ); \
    p##_1 = vec_perm( p##_0, w, perm ); /* a.c2 b.c2 */ \
    perm = vec_gpci( 00167 ); \
    p##_0 = vec_perm( p##_0, p##_2, perm ); /* a.c1 b.c1 */ \
        \
    p##_2 = VEC_LOAD2(0l,&a.c3.re); \
    p##_2 = vec_perm( p##_2, w, perm ) /* a.c3 b.c3 */

#define _qpx_vectorpair_ua_load(p,a,b,w,perm) \
    p##_0 = VEC_LOAD(0l,&b.c1.re); \
    p##_2 = VEC_LOAD2(0l,&a.c1.re); \
    w = VEC_LOAD(0l,&a.c2.re); \
        \
    perm = vec_gpci( 00167 ); \
    p##_1 = vec_perm( w, p##_0, perm ); /* a.c2 b.c2 */ \
    perm = vec_gpci( 02345 ); \
    p##_0 = vec_perm( p##_2, p##_0, perm ); /* a.c1 b.c1 */ \
        \
    p##_2 = VEC_LOAD2(0l,&b.c3.re); \
    p##_2 = vec_perm( w, p##_2, perm ) /* a.c3 b.c3 */

#define _qpx_vectorpair_weyl_load(p,a,w,perm) \
    p##_0 = VEC_LOAD(0l,&a.c1.c1.re); \
    w = VEC_LOAD(0l,&a.c1.c3.re); \
    p##_2 = VEC_LOAD(0l,&a.c2.c2.re); \
        \
    perm = vec_gpci( 02345 ); \
    p##_1 = vec_perm( p##_0, p##_2, perm ); /* s1.c2 s2.c2 */ \
    perm = vec_gpci( 00167 ); \
    p##_0 = vec_perm( p##_0, w, perm ); /* s1.c1 s2.c1 */ \
    p##_2 = vec_perm( w, p##_2, perm ) /* s1.c3 s2.c3 */

#define _qpx_vectorpair_weyl34_load(p,a,w,perm) \
    p##_0 = VEC_LOAD(0l,&a.c3.c1.re); \
    w = VEC_LOAD(0l,&a.c3.c3.re); \
    p##_2 = VEC_LOAD(0l,&a.c4.c2.re); \
        \
    perm = vec_gpci( 02345 ); \
    p##_1 = vec_perm( p##_0, p##_2, perm ); /* s1.c2 s2.c2 */ \
    perm = vec_gpci( 00167 ); \
    p##_0 = vec_perm( p##_0, w, perm ); /* s1.c1 s2.c1 */ \
    p##_2 = vec_perm( w, p##_2, perm ) /* s1.c3 s2.c3 */

#define _qpx_vectorpair_spinor34_load(p1,p2,s,w0,w1,perm) \
    p1##_0 = VEC_LOAD( 0l, &s.c1.c1.re ); /* s1.c1 s1.c2 */ \
    w0     = VEC_LOAD( 0l, &s.c1.c3.re ); /* s1.c3 s2.c1 */ \
    p1##_2 = VEC_LOAD( 0l, &s.c2.c2.re ); /* s2.c2 s2.c3 */ \
    p2##_0 = VEC_LOAD( 0l, &s.c3.c1.re ); /* s3.c1 s3.c2 */ \
    w1     = VEC_LOAD( 0l, &s.c3.c3.re ); /* s3.c3 s4.c1 */ \
    p2##_2 = VEC_LOAD( 0l, &s.c4.c2.re ); /* s4.c2 s4.c3 */ \
        \
    perm = vec_gpci( 02345 ); \
    p1##_1 = vec_perm( p1##_0, p1##_2, perm ); /* s1.c2 s2.c2 */ \
    p2##_1 = vec_perm( p2##_0, p2##_2, perm ); /* s3.c2 s4.c2 */ \
    perm = vec_gpci( 00167 ); \
    p1##_0 = vec_perm( p1##_0, w0, perm ); /* s1.c1 s2.c1 */ \
    p2##_0 = vec_perm( p2##_0, w1, perm ); /* s3.c1 s4.c1 */ \
    p1##_2 = vec_perm( w0, p1##_2, perm ); /* s1.c3 s2.c3 */ \
    p2##_2 = vec_perm( w1, p2##_2, perm ) /* s3.c3 s4.c3 */ 

#define _qpx_vectorpair_spinor43_load(p1,p2,s,w0,w1,perm_2345,perm_0167) \
    p1##_0 = VEC_LOAD( 0l, &s.c1.c1.re ); /* s1.c1 s1.c2 */ \
    w0     = VEC_LOAD( 0l, &s.c1.c3.re ); /* s1.c3 s2.c1 */ \
    p1##_2 = VEC_LOAD( 0l, &s.c2.c2.re ); /* s2.c2 s2.c3 */ \
    p2##_0 = VEC_LOAD( 0l, &s.c3.c1.re ); /* s3.c1 s3.c2 */ \
    w1     = VEC_LOAD( 0l, &s.c3.c3.re ); /* s3.c3 s4.c1 */ \
    p2##_2 = VEC_LOAD( 0l, &s.c4.c2.re ); /* s4.c2 s4.c3 */ \
        \
    perm_2345 = vec_gpci( 02345 ); \
    p1##_1 = vec_perm( p1##_0, p1##_2, perm_2345 ); /* s1.c2 s2.c2 */ \
    perm_0167 = vec_gpci( 00167 ); \
    p2##_1 = vec_perm( p2##_2, p2##_0, perm_0167 ); /* s4.c2 s3.c2 */ \
    p1##_0 = vec_perm( p1##_0, w0,     perm_0167 ); /* s1.c1 s2.c1 */ \
    p2##_0 = vec_perm( w1,     p2##_0, perm_2345 ); /* s4.c1 s3.c1 */ \
    p1##_2 = vec_perm( w0,     p1##_2, perm_0167 ); /* s1.c3 s2.c3 */ \
    p2##_2 = vec_perm( p2##_2, w1,     perm_2345 ) /* s4.c3 s3.c3 */ 

/*  store   */

#if 0
#define _qpx_vectorpair_aa_store(rs1,rs2,p,w1,w2) \
    w1 = vec_gpci( 00145 ); \
    w1 = vec_perm( p##_0, p##_1, w1 ); \
        \
    w2 = vec_gpci( 02367 ); \
    p##_0 = vec_perm( p##_0, p##_1, w2 ); \
    w2 = vec_perm( p##_2, p##_2, w2 ); \
        \
    VEC_STORE(w1,0l,&(rs1.c1.re)); \
    VEC_STORE2(##p##_2,0l,&(rs1.c3.re)); \
    VEC_STORE(##p##_0,0l,&(rs2.c1.re)); \
    VEC_STORE2(w2,0l,&(rs2.c3.re))
#endif

#define _qpx_vectorpair_au_store(rs1,rs2,p,w1,w2) \
    w1 = vec_gpci( 00145 ); \
    w1 = vec_perm( p##_0, p##_1, w1 ); \
        \
    w2 = vec_gpci( 02367 ); \
    p##_0 = vec_perm( p##_0, p##_0, w2 ); \
    w2 = vec_perm( p##_1, p##_2, w2 ); \
        \
    VEC_STORE(w1,0l,&(rs1.c1.re)); \
    VEC_STORE2(##p##_2,0l,&(rs1.c3.re)); \
    VEC_STORE2(##p##_0,0l,&(rs2.c1.re)); \
    VEC_STORE(w2,0l,&(rs2.c2.re))

#define _qpx_vectorpair_weyl_store(rs,p,wvec,perm) \
    perm = vec_gpci( 00145 ); \
    wvec##_0 = vec_perm( p##_0, p##_1, perm ); /* s1.c1 s1.c2 */ \
    perm = vec_gpci( 00167 ); \
    wvec##_1 = vec_perm( p##_2, p##_0, perm ); /* s1.c3 s2.c1 */ \
    perm = vec_gpci( 02367 ); \
    wvec##_2 = vec_perm( p##_1, p##_2, perm ); /* s2.c2 s2.c3 */ \
        \
    VEC_STORE(wvec##_0, 0l, &rs.c1.c1.re ); \
    VEC_STORE(wvec##_1, 0l, &rs.c1.c3.re ); \
    VEC_STORE(wvec##_2, 0l, &rs.c2.c2.re )

#define _qpx_vectorpair_weyl34_store(rs,p,wvec,perm) \
    perm = vec_gpci( 00145 ); \
    wvec##_0 = vec_perm( p##_0, p##_1, perm ); /* s1.c1 s1.c2 */ \
    perm = vec_gpci( 00167 ); \
    wvec##_1 = vec_perm( p##_2, p##_0, perm ); /* s1.c3 s2.c1 */ \
    perm = vec_gpci( 02367 ); \
    wvec##_2 = vec_perm( p##_1, p##_2, perm ); /* s2.c2 s2.c3 */ \
        \
    VEC_STORE(wvec##_0, 0l, &rs.c3.c1.re ); \
    VEC_STORE(wvec##_1, 0l, &rs.c3.c3.re ); \
    VEC_STORE(wvec##_2, 0l, &rs.c4.c2.re )

#define _qpx_vectorpair_spinor34_store(s,p1,p2,w0,w1,perm) \
    perm = vec_gpci( 00167 ); \
    w0 = vec_perm( p1##_2, p1##_0, perm ); /* s1.c3 s2.c1 */ \
    w1 = vec_perm( p2##_2, p2##_0, perm ); /* s3.c3 s4.c1 */ \
    perm = vec_gpci( 00145 ); \
    p1##_0 = vec_perm( p1##_0, p1##_1, perm ); /* s1.c1 s1.c2 */ \
    p2##_0 = vec_perm( p2##_0, p2##_1, perm ); /* s3.c1 s3.c2 */ \
    perm = vec_gpci( 02367 ); \
    p1##_2 = vec_perm( p1##_1, p1##_2, perm ); /* s2.c2 s2.c3 */ \
    p2##_2 = vec_perm( p2##_1, p2##_2, perm ); /* s4.c2 s4.c3 */ \
        \
    VEC_STORE(p1##_0, 0l, &s.c1.c1.re ); \
    VEC_STORE(w0,     0l, &s.c1.c3.re ); \
    VEC_STORE(p1##_2, 0l, &s.c2.c2.re ); \
    VEC_STORE(p2##_0, 0l, &s.c3.c1.re ); \
    VEC_STORE(w1,     0l, &s.c3.c3.re ); \
    VEC_STORE(p2##_2, 0l, &s.c4.c2.re )

/*  assign  */

#define _qpx_vectorpair_assign(lhs,rhs) \
    lhs##_0 = rhs##_0; \
    lhs##_1 = rhs##_1; \
    lhs##_2 = rhs##_2 \

/*  arithmetics */

#define _qpx_vectorpair_mul(pr,pa,w) \
    pr##_0 = vec_mul( pa##_0, w ); \
    pr##_1 = vec_mul( pa##_1, w ); \
    pr##_2 = vec_mul( pa##_2, w )

#define _qpx_vectorpair_add(pr,pa,pb) \
    pr##_0 = vec_add( pa##_0, pb##_0 ); \
    pr##_1 = vec_add( pa##_1, pb##_1 ); \
    pr##_2 = vec_add( pa##_2, pb##_2 )

#define _qpx_vectorpair_sub(pr,pa,pb) \
    pr##_0 = vec_sub( pa##_0, pb##_0 ); \
    pr##_1 = vec_sub( pa##_1, pb##_1 ); \
    pr##_2 = vec_sub( pa##_2, pb##_2 )

#define _qpx_vectorpair_1add2sub(pr,pa,pb,w) \
    w = vec_splats( 1. ); \
    w[2] = w[3] = -1.; \
        \
    pr##_0 = vec_madd( pb##_0, w, pa##_0 ); \
    pr##_1 = vec_madd( pb##_1, w, pa##_1 ); \
    pr##_2 = vec_madd( pb##_2, w, pa##_2 )

#define _qpx_vectorpair_1sub2add(pr,pa,pb,w) \
    w = vec_splats( -1. ); \
    w[2] = w[3] = 1.; \
        \
    pr##_0 = vec_madd( pb##_0, w, pa##_0 ); \
    pr##_1 = vec_madd( pb##_1, w, pa##_1 ); \
    pr##_2 = vec_madd( pb##_2, w, pa##_2 )

#define _qpx_vectorpair_x1add2sub(pr,pa,pb,wvec,perm) \
    perm = vec_gpci( 02301 ); \
    wvec##_0 = vec_perm( pb##_0, pb##_0, perm ); \
    wvec##_1 = vec_perm( pb##_1, pb##_1, perm ); \
    wvec##_2 = vec_perm( pb##_2, pb##_2, perm ); \
        \
    _qpx_vectorpair_1add2sub(pr,pa,wvec,perm)

#define _qpx_vectorpair_x1sub2add(pr,pa,pb,wvec,perm) \
    perm = vec_gpci( 02301 ); \
    wvec##_0 = vec_perm( pb##_0, pb##_0, perm ); \
    wvec##_1 = vec_perm( pb##_1, pb##_1, perm ); \
    wvec##_2 = vec_perm( pb##_2, pb##_2, perm ); \
        \
    _qpx_vectorpair_1sub2add(pr,pa,wvec,perm)

#define _qpx_vectorpair_i_add(pr,pa,pb,i) \
    i = vec_splats( 0. ); \
    i[1] = i[3] = 1.; \
        \
    pr##_0 = vec_xxnpmadd( pb##_0, i, pa##_0 ); \
    pr##_1 = vec_xxnpmadd( pb##_1, i, pa##_1 ); \
    pr##_2 = vec_xxnpmadd( pb##_2, i, pa##_2 )

#define _qpx_vectorpair_i_sub(pr,pa,pb,i) \
    i = vec_splats( 0. ); \
    i[1] = i[3] = 1.; \
        \
    pr##_0 = vec_xxcpnmadd( pb##_0, i, pa##_0 ); \
    pr##_1 = vec_xxcpnmadd( pb##_1, i, pa##_1 ); \
    pr##_2 = vec_xxcpnmadd( pb##_2, i, pa##_2 )

#define _qpx_vectorpair_i_1add2sub(pr,pa,pb,i) \
    i = vec_splats( 0. ); \
    i[1] = 1; \
    i[3] = -1.; \
        \
    pr##_0 = vec_xxnpmadd( pb##_0, i, pa##_0 ); \
    pr##_1 = vec_xxnpmadd( pb##_1, i, pa##_1 ); \
    pr##_2 = vec_xxnpmadd( pb##_2, i, pa##_2 )

#define _qpx_vectorpair_i_1sub2add(pr,pa,pb,i) \
    i = vec_splats( 0. ); \
    i[1] = 1; \
    i[3] = -1.; \
        \
    pr##_0 = vec_xxcpnmadd( pb##_0, i, pa##_0 ); \
    pr##_1 = vec_xxcpnmadd( pb##_1, i, pa##_1 ); \
    pr##_2 = vec_xxcpnmadd( pb##_2, i, pa##_2 )

#define _qpx_vectorpair_i_xadd(pr,pa,pb,wvec,perm) \
    perm = vec_gpci( 02301 ); \
    wvec##_0 = vec_perm( pb##_0, pb##_0, perm ); \
    wvec##_1 = vec_perm( pb##_1, pb##_1, perm ); \
    wvec##_2 = vec_perm( pb##_2, pb##_2, perm ); \
        \
    _qpx_vectorpair_i_add(pr,pa,wvec,perm)

#define _qpx_vectorpair_i_xsub(pr,pa,pb,wvec,perm) \
    perm = vec_gpci( 02301 ); \
    wvec##_0 = vec_perm( pb##_0, pb##_0, perm ); \
    wvec##_1 = vec_perm( pb##_1, pb##_1, perm ); \
    wvec##_2 = vec_perm( pb##_2, pb##_2, perm ); \
        \
    _qpx_vectorpair_i_sub(pr,pa,wvec,perm)

/*
 * su3 matrix functions
 *
 * */

#define _qpx_su3_declare(cm) \
    vector4double cm##_0_0, cm##_0_1, cm##_0_2, \
                cm##_1_0, cm##_1_1, cm##_1_2, \
                cm##_2_0, cm##_2_1, cm##_2_2 

#define _qpx_su3_save_load(cm,u) \
    cm##_0_0 = VEC_LOAD2(0l,&u.c11.re); \
    cm##_0_1 = VEC_LOAD2(0l,&u.c12.re); \
    cm##_0_2 = VEC_LOAD2(0l,&u.c13.re); \
    cm##_1_0 = VEC_LOAD2(0l,&u.c21.re); \
    cm##_1_1 = VEC_LOAD2(0l,&u.c22.re); \
    cm##_1_2 = VEC_LOAD2(0l,&u.c23.re); \
    cm##_2_0 = VEC_LOAD2(0l,&u.c31.re); \
    cm##_2_1 = VEC_LOAD2(0l,&u.c32.re); \
    cm##_2_2 = VEC_LOAD2(0l,&u.c33.re) 

#if 1
#define _qpx_su3_a_load(cm,u,perm) \
    cm##_0_0 = VEC_LOAD(0l,&u.c11.re); \
    cm##_0_2 = VEC_LOAD(0l,&u.c13.re); \
    cm##_1_1 = VEC_LOAD(0l,&u.c22.re); \
    cm##_2_0 = VEC_LOAD(0l,&u.c31.re); \
    cm##_2_2 = VEC_LOAD2(0l,&u.c33.re); \
        \
    perm = vec_gpci( 02323 ); \
    cm##_0_1 = vec_perm( cm##_0_0, cm##_0_0, perm ); \
    cm##_1_0 = vec_perm( cm##_0_2, cm##_0_2, perm ); \
    cm##_1_2 = vec_perm( cm##_1_1, cm##_1_1, perm ); \
    cm##_2_1 = vec_perm( cm##_2_0, cm##_2_0, perm ); \
        \
    perm = vec_gpci( 00101 ); \
    cm##_0_0 = vec_perm( cm##_0_0, cm##_0_0, perm ); \
    cm##_0_2 = vec_perm( cm##_0_2, cm##_0_2, perm ); \
    cm##_1_1 = vec_perm( cm##_1_1, cm##_1_1, perm ); \
    cm##_2_0 = vec_perm( cm##_2_0, cm##_2_0, perm )

#define _qpx_su3_u_load(cm,u,perm) \
    cm##_0_0 = VEC_LOAD2(0l,&u.c11.re); \
    cm##_0_1 = VEC_LOAD(0l,&u.c12.re); \
    cm##_1_0 = VEC_LOAD(0l,&u.c21.re); \
    cm##_1_2 = VEC_LOAD(0l,&u.c23.re); \
    cm##_2_1 = VEC_LOAD(0l,&u.c32.re); \
        \
    perm = vec_gpci( 02323 ); \
    cm##_0_2 = vec_perm( cm##_0_1, cm##_0_1, perm ); \
    cm##_1_1 = vec_perm( cm##_1_0, cm##_1_0, perm ); \
    cm##_2_0 = vec_perm( cm##_1_2, cm##_1_2, perm ); \
    cm##_2_2 = vec_perm( cm##_2_1, cm##_2_1, perm ); \
        \
    perm = vec_gpci( 00101 ); \
    cm##_0_1 = vec_perm( cm##_0_1, cm##_0_1, perm ); \
    cm##_1_0 = vec_perm( cm##_1_0, cm##_1_0, perm ); \
    cm##_1_2 = vec_perm( cm##_1_2, cm##_1_2, perm ); \
    cm##_2_1 = vec_perm( cm##_2_1, cm##_2_1, perm )

#else
#define _qpx_su3_a_load(cm,u,perm) \
    _qpx_su3_save_load(cm,u)

#define _qpx_su3_u_load(cm,u,perm) \
    _qpx_su3_save_load(cm,u)

#endif // switch su3 load

#define _qpx_su3_multiply(rp,cm,p) \
    rp##_0 = vec_xmul( p##_0, cm##_0_0 ); \
    rp##_0 = vec_xxnpmadd( cm##_0_0, p##_0, rp##_0 ); \
    rp##_0 = vec_xmadd( p##_1, cm##_0_1, rp##_0 ); \
    rp##_0 = vec_xxnpmadd( cm##_0_1, p##_1, rp##_0 ); \
    rp##_0 = vec_xmadd( p##_2, cm##_0_2, rp##_0 ); \
    rp##_0 = vec_xxnpmadd( cm##_0_2, p##_2, rp##_0 ); \
        \
    rp##_1 = vec_xmul( p##_0, cm##_1_0 ); \
    rp##_1 = vec_xxnpmadd( cm##_1_0, p##_0, rp##_1 ); \
    rp##_1 = vec_xmadd( p##_1, cm##_1_1, rp##_1 ); \
    rp##_1 = vec_xxnpmadd( cm##_1_1, p##_1, rp##_1 ); \
    rp##_1 = vec_xmadd( p##_2, cm##_1_2, rp##_1 ); \
    rp##_1 = vec_xxnpmadd( cm##_1_2, p##_2, rp##_1 ); \
        \
    rp##_2 = vec_xmul( p##_0, cm##_2_0 ); \
    rp##_2 = vec_xxnpmadd( cm##_2_0, p##_0, rp##_2 ); \
    rp##_2 = vec_xmadd( p##_1, cm##_2_1, rp##_2 ); \
    rp##_2 = vec_xxnpmadd( cm##_2_1, p##_1, rp##_2 ); \
    rp##_2 = vec_xmadd( p##_2, cm##_2_2, rp##_2 ); \
    rp##_2 = vec_xxnpmadd( cm##_2_2, p##_2, rp##_2 ) 

#define _qpx_su3_inv_multiply(rp,u,p) \
    rp##_0 = vec_xmul( cm##_0_0, p##_0 ); \
    rp##_0 = vec_xxcpnmadd( p##_0, cm##_0_0, rp##_0 ); \
    rp##_0 = vec_xmadd( cm##_1_0, p##_1, rp##_0 ); \
    rp##_0 = vec_xxcpnmadd( p##_1, cm##_1_0, rp##_0 ); \
    rp##_0 = vec_xmadd( cm##_2_0, p##_2, rp##_0 ); \
    rp##_0 = vec_xxcpnmadd( p##_2, cm##_2_0, rp##_0 ); \
        \
    rp##_1 = vec_xmul( cm##_0_1, p##_0 ); \
    rp##_1 = vec_xxcpnmadd( p##_0, cm##_0_1, rp##_1 ); \
    rp##_1 = vec_xmadd( cm##_1_1, p##_1, rp##_1 ); \
    rp##_1 = vec_xxcpnmadd( p##_1, cm##_1_1, rp##_1 ); \
    rp##_1 = vec_xmadd( cm##_2_1, p##_2, rp##_1 ); \
    rp##_1 = vec_xxcpnmadd( p##_2, cm##_2_1, rp##_1 ); \
        \
    rp##_2 = vec_xmul( cm##_0_2, p##_0 ); \
    rp##_2 = vec_xxcpnmadd( p##_0, cm##_0_2, rp##_2 ); \
    rp##_2 = vec_xmadd( cm##_1_2, p##_1, rp##_2 ); \
    rp##_2 = vec_xxcpnmadd( p##_1, cm##_1_2, rp##_2 ); \
    rp##_2 = vec_xmadd( cm##_2_2, p##_2, rp##_2 ); \
    rp##_2 = vec_xxcpnmadd( p##_2, cm##_2_2, rp##_2 )

/*
 * pauli matrix functions
 *
 * */

#define _qpx_pauli_declare(pm) \
    vector4double pm##_0, pm##_1, pm##_2, \
                pm##_3, pm##_4, pm##_5, \
                pm##_6, pm##_7, pm##_8 

#define _qpx_pauli_load_all(m,p) \
    m##_0 = VEC_LOAD(0l, &p.u[0] ); \
    m##_1 = VEC_LOAD(0l, &p.u[4] ); \
    m##_2 = VEC_LOAD(0l, &p.u[8] ); \
    m##_3 = VEC_LOAD(0l, &p.u[12] ); \
    m##_4 = VEC_LOAD(0l, &p.u[16] ); \
    m##_5 = VEC_LOAD(0l, &p.u[20] ); \
    m##_6 = VEC_LOAD(0l, &p.u[24] ); \
    m##_7 = VEC_LOAD(0l, &p.u[28] ); \
    m##_8 = VEC_LOAD(0l, &p.u[32] ) 

#define _qpx_pauli_multiply_all(rp,m,p, perm,w, perm_2301,fact_pppm) \
        /* spin diagonal - color diagonal (real) */ \
    perm = vec_gpci( 00033 ); \
    w = vec_perm( m##_0, m##_0, perm ); /* 0,0    3,3 */ \
    rp##_0 = vec_mul( p##_0, w ); \
    perm = vec_gpci( 01144 ); \
    w = vec_perm( m##_0, m##_1, perm ); /* 1,1    4,4 */ \
    rp##_1 = vec_mul( p##_1, w ); \
    perm = vec_gpci( 02255 ); \
    w = vec_perm( m##_0, m##_1, perm ); /* 2,2    5,5 */ \
    rp##_2 = vec_mul( p##_2, w ); \
        \
        /* spin diagonal - color mixing */ \
    perm = vec_gpci( 02367 ); \
    w = vec_perm( m##_1, m##_7, perm ); /* 6,7    30,31 */ \
    rp##_0 = vec_xmadd( w, p##_1,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd( p##_1, w, rp##_0 ); \
    rp##_1 = vec_xmadd( w, p##_0,    rp##_1 ); \
    rp##_1 = vec_xxcpnmadd(p##_0, w, rp##_1 ); \
        \
    perm = vec_gpci( 00145 ); \
    w = vec_perm( m##_2, m##_8, perm ); /* 8,9    32,33 */ \
    rp##_0 = vec_xmadd( w, p##_2,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd( p##_2, w, rp##_0 ); \
    rp##_2 = vec_xmadd( w, p##_0,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd(p##_0, w, rp##_2 ); \
        \
    perm = vec_gpci( 00167 ); \
    w = vec_perm( m##_4, m##_8, perm ); /* 16,17  34,35 */ \
    rp##_1 = vec_xmadd( w, p##_2,    rp##_1 ); \
    rp##_1 = vec_xxnpmadd( p##_2, w, rp##_1 ); \
    rp##_2 = vec_xmadd( w, p##_1,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd(p##_1, w, rp##_2 ); \
        \
        /* spin mixing (switch input) */ \
    perm_2301 = vec_gpci( 02301 ); \
    p##_0 = vec_perm( p##_0, p##_0, perm_2301 ); \
    p##_1 = vec_perm( p##_1, p##_1, perm_2301 ); \
    p##_2 = vec_perm( p##_2, p##_2, perm_2301 ); \
        \
    fact_pppm = vec_splats( 1. ); \
    fact_pppm[3] = -1.; \
        \
        /* spin mixing - color diagonal */ \
    perm = vec_gpci( 02323 ); \
    w = vec_perm( m##_2, m##_2, perm ); \
    w = vec_mul( w, fact_pppm );    /* 10,11    10,-11 */ \
    rp##_0 = vec_xmadd( w,  p##_0,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd(  p##_0, w, rp##_0 ); \
        \
    perm = vec_gpci( 00101 ); \
    w = vec_perm( m##_5, m##_5, perm ); \
    w = vec_mul( w, fact_pppm );    /* 20,21    20,-21 */ \
    rp##_1 = vec_xmadd( w,  p##_1,    rp##_1 ); \
    rp##_1 = vec_xxnpmadd(  p##_1, w, rp##_1 ); \
        \
    w = vec_perm( m##_7, m##_7, perm ); \
    w = vec_mul( w, fact_pppm );    /* 28,29    28,-29 */ \
    rp##_2 = vec_xmadd( w,  p##_2,    rp##_2 ); \
    rp##_2 = vec_xxnpmadd(  p##_2, w, rp##_2 ); \
        \
        /* spin mixing - color mixing */ \
    perm = vec_gpci( 00167 ); \
    w = vec_perm( m##_3, m##_4, perm ); \
    w = vec_mul( w, fact_pppm );    /* 12,13    18,-19 */ \
    rp##_0 = vec_xmadd( w,  p##_1,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd(  p##_1, w, rp##_0 ); \
    w = vec_perm( w, w, perm_2301 ); \
    rp##_1 = vec_xmadd( w,  p##_0,    rp##_1 ); \
    rp##_1 = vec_xxcpnmadd( p##_0, w, rp##_1 ); \
        \
    perm = vec_gpci( 02345 ); \
    w = vec_perm( m##_3, m##_6, perm ); \
    w = vec_mul( w, fact_pppm );    /* 14,15    24,-25 */ \
    rp##_0 = vec_xmadd( w,  p##_2,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd(  p##_2, w, rp##_0 ); \
    w = vec_perm( w, w, perm_2301 ); \
    rp##_2 = vec_xmadd( w,  p##_0,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd( p##_0, w, rp##_2 ); \
        \
    perm = vec_gpci( 02367 ); \
    w = vec_perm( m##_5, m##_6, perm ); \
    w = vec_mul( w, fact_pppm );    /* 22,23    26,-27 */ \
    rp##_1 = vec_xmadd( w,  p##_2,    rp##_1 ); \
    rp##_1 = vec_xxnpmadd(  p##_2, w, rp##_1 ); \
    w = vec_perm( w, w, perm_2301 ); \
    rp##_2 = vec_xmadd( w,  p##_1,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd( p##_1, w, rp##_2 ); \
        \
        /* (switch input back) */ \
    p##_0 = vec_perm( p##_0, p##_0, perm_2301 ); \
    p##_1 = vec_perm( p##_1, p##_1, perm_2301 ); \
    p##_2 = vec_perm( p##_2, p##_2, perm_2301 ) 
    

/* CAUTION
 * this macros gives the correct result,
 * but it will modify the matrix and the initial
 * original vector:
 * - the matrix first 8 elements will be destroyed
 * - the vector will be crossed!
 * */

#define _qpx_pauli_multiply_unsafe(rp,m,p, perm,w) \
        /* spin diagonal - color diagonal (real) */ \
    perm = vec_gpci( 00033 ); \
    w = vec_perm( m##_0, m##_0, perm ); /* 0,0    3,3 */ \
    rp##_0 = vec_mul( p##_0, w ); \
    perm = vec_gpci( 01144 ); \
    w = vec_perm( m##_0, m##_1, perm ); /* 1,1    4,4 */ \
    rp##_1 = vec_mul( p##_1, w ); \
    perm = vec_gpci( 02255 ); \
    w = vec_perm( m##_0, m##_1, perm ); /* 2,2    5,5 */ \
    rp##_2 = vec_mul( p##_2, w ); \
        \
        /* spin diagonal - color mixing */ \
    perm = vec_gpci( 02367 ); \
    w = vec_perm( m##_1, m##_7, perm ); /* 6,7    30,31 */ \
    rp##_0 = vec_xmadd( w, p##_1,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd( p##_1, w, rp##_0 ); \
    rp##_1 = vec_xmadd( w, p##_0,    rp##_1 ); \
    rp##_1 = vec_xxcpnmadd(p##_0, w, rp##_1 ); \
        \
    perm = vec_gpci( 00145 ); \
    w = vec_perm( m##_2, m##_8, perm ); /* 8,9    32,33 */ \
    rp##_0 = vec_xmadd( w, p##_2,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd( p##_2, w, rp##_0 ); \
    rp##_2 = vec_xmadd( w, p##_0,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd(p##_0, w, rp##_2 ); \
        \
    perm = vec_gpci( 00167 ); \
    w = vec_perm( m##_4, m##_8, perm ); /* 16,17  34,35 */ \
    rp##_1 = vec_xmadd( w, p##_2,    rp##_1 ); \
    rp##_1 = vec_xxnpmadd( p##_2, w, rp##_1 ); \
    rp##_2 = vec_xmadd( w, p##_1,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd(p##_1, w, rp##_2 ); \
        \
        /* spin mixing (switch input) */ \
    m##_0 = vec_gpci( 02301 ); \
    p##_0 = vec_perm( p##_0, p##_0, m##_0 ); \
    p##_1 = vec_perm( p##_1, p##_1, m##_0 ); \
    p##_2 = vec_perm( p##_2, p##_2, m##_0 ); \
        \
    m##_1 = vec_splats( 1. ); \
    m##_1[3] = -1.; \
        \
        /* spin mixing - color diagonal */ \
    perm = vec_gpci( 02323 ); \
    w = vec_perm( m##_2, m##_2, perm ); \
    w = vec_mul( w, m##_1 );    /* 10,11    10,-11 */ \
    rp##_0 = vec_xmadd( w,  p##_0,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd(  p##_0, w, rp##_0 ); \
        \
    perm = vec_gpci( 00101 ); \
    w = vec_perm( m##_5, m##_5, perm ); \
    w = vec_mul( w, m##_1 );    /* 20,21    20,-21 */ \
    rp##_1 = vec_xmadd( w,  p##_1,    rp##_1 ); \
    rp##_1 = vec_xxnpmadd(  p##_1, w, rp##_1 ); \
        \
    w = vec_perm( m##_7, m##_7, perm ); \
    w = vec_mul( w, m##_1 );    /* 28,29    28,-29 */ \
    rp##_2 = vec_xmadd( w,  p##_2,    rp##_2 ); \
    rp##_2 = vec_xxnpmadd(  p##_2, w, rp##_2 ); \
        \
        /* spin mixing - color mixing */ \
    perm = vec_gpci( 00167 ); \
    w = vec_perm( m##_3, m##_4, perm ); \
    w = vec_mul( w, m##_1 );    /* 12,13    18,-19 */ \
    rp##_0 = vec_xmadd( w,  p##_1,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd(  p##_1, w, rp##_0 ); \
    w = vec_perm( w, w, m##_0 ); \
    rp##_1 = vec_xmadd( w,  p##_0,    rp##_1 ); \
    rp##_1 = vec_xxcpnmadd( p##_0, w, rp##_1 ); \
        \
    perm = vec_gpci( 02345 ); \
    w = vec_perm( m##_3, m##_6, perm ); \
    w = vec_mul( w, m##_1 );    /* 14,15    24,-25 */ \
    rp##_0 = vec_xmadd( w,  p##_2,    rp##_0 ); \
    rp##_0 = vec_xxnpmadd(  p##_2, w, rp##_0 ); \
    w = vec_perm( w, w, m##_0 ); \
    rp##_2 = vec_xmadd( w,  p##_0,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd( p##_0, w, rp##_2 ); \
        \
    perm = vec_gpci( 02367 ); \
    w = vec_perm( m##_5, m##_6, perm ); \
    w = vec_mul( w, m##_1 );    /* 22,23    26,-27 */ \
    rp##_1 = vec_xmadd( w,  p##_2,    rp##_1 ); \
    rp##_1 = vec_xxnpmadd(  p##_2, w, rp##_1 ); \
    w = vec_perm( w, w, m##_0 ); \
    rp##_2 = vec_xmadd( w,  p##_1,    rp##_2 ); \
    rp##_2 = vec_xxcpnmadd( p##_1, w, rp##_2 ) 

#endif // QPX_H
