!===============================================================================
!
! module_function_decl.F90  -  declaration of functions
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2003-2016 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_function_decl

  !----------------------------------------------------------------------------
  ! functions in misc.F90:
  !----------------------------------------------------------------------------

  character(len=10), external :: datum
  character(len=12), external :: uhrzeit
  character(len=20), external :: rechner
  logical, external :: f_exist
  character(len=12), external :: itoa
  character(len=24), external :: i8toa
  character(len=24), external :: REALtoa
  integer, external :: getdigit  ! returns a digit: length of a given integer in characters 

  !----------------------------------------------------------------------------
  ! processes I (-> comm/pes*.F90):
  !----------------------------------------------------------------------------

  integer, external :: num_pes

!!#ifdef CRAY
!!  integer, intrinsic :: my_pe  ! gives the same id as MPI_Rank in MPI_COMM_WORLD
!!#else
  integer, external :: my_pe
!!#endif

  !----------------------------------------------------------------------------
  ! processes II (-> index2.F90)
  !----------------------------------------------------------------------------

  integer, external :: coord2pe

  !----------------------------------------------------------------------------
  ! global reduction (-> comm/reduction_*.F90)
  !----------------------------------------------------------------------------

  REAL, external :: dotprod
  REAL, external :: global_sum
  REAL, external :: global_min
  REAL, external :: global_max

  !----------------------------------------------------------------------------
  ! sc-field (-> sc.F90)
  !----------------------------------------------------------------------------

  REAL, external    :: sc_norm2
  REAL, external    :: sc_dot
  COMPLEX, external :: sc_cdotc

  !----------------------------------------------------------------------------
  ! sc-field (-> sc_r4.F90)
  !----------------------------------------------------------------------------

  REAL4, external    :: sc_norm2_r4
  REAL4, external    :: sc_dot_r4
  COMPLEX4, external :: sc_cdotc_r4

  !----------------------------------------------------------------------------
  ! sc-field (-> sc_r16.F90)
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! random numbers (-> ran/.../ran.F90):
  !----------------------------------------------------------------------------

  RAN_NAME, external :: ran_name
  REAL, external     :: ran_number
  SEED, external     :: ran_get_seed

  !----------------------------------------------------------------------------
  ! identification of D (-> d/DVersion.F90)
  !----------------------------------------------------------------------------

  integer, external :: version_of_d

  !----------------------------------------------------------------------------
  ! communication method (-> comm/comm_*.F90)
  !----------------------------------------------------------------------------

  COMM_METHOD, external :: comm_method


contains
  !----------------------------------------------------------------------------
  ! functions for thread management 
  !----------------------------------------------------------------------------

#ifdef _OPENMP
  !----------------------------------------------------------------------------
  ! wrappers for OpenMP runtime functions
  !----------------------------------------------------------------------------
  integer function my_thread()
    use omp_lib
    implicit none
    my_thread = omp_get_thread_num()
  end function my_thread

  !----------------------------------------------------------------------------
  integer function num_threads()  ! returns the value of OMP_NUM_THREADS
    use omp_lib
    implicit none
    num_threads = omp_get_max_threads()
  end function num_threads

  !----------------------------------------------------------------------------
  integer function active_threads()  ! size of the current team
    use omp_lib
    implicit none
    active_threads = omp_get_num_threads()
  end function active_threads

  !----------------------------------------------------------------------------
  integer function first_thread()  ! used for overlapping commuication and computation
    use omp_lib
    implicit none
    if (omp_get_num_threads() > 1) then  ! if there is more that 1 active thread then
       first_thread = 1                  !    threads >= first_thread compute
    else                                 ! else
       first_thread = 0                  !    all threads compute
    endif
  end function first_thread

#else
  !----------------------------------------------------------------------------
  ! stubs
  !----------------------------------------------------------------------------
  integer function my_thread()
    implicit none
    my_thread = 0
  end function my_thread

  !----------------------------------------------------------------------------
  integer function num_threads()
    implicit none
    num_threads = 1
  end function num_threads

  !----------------------------------------------------------------------------
  integer function active_threads()
    implicit none
    active_threads = 1
  end function active_threads

  !----------------------------------------------------------------------------
  integer function first_thread()
    implicit none
    first_thread = 0
  end function first_thread
#endif

end

!===============================================================================
